import dataclasses
import datetime
import json
import uuid as uuidpkg  # to avoid conflict with uuid type annotation below

from dateutil.parser import isoparse

# ZeroTime is the zero time used for datetime.datetime types.
ZeroTime: datetime.datetime = datetime.datetime(
    1, 1, 1, 0, 0, tzinfo=datetime.timezone.utc
)

# ZeroUUID is the zero UUID used for UUID types.
ZeroUUID: uuidpkg.UUID = uuidpkg.UUID(int=0)

# A helper enum type; see values below.
LogEntryType = str

# Various values for LogEntryType
LogEntryUnknown: LogEntryType = ""
LogEntryLog: LogEntryType = "log"
LogEntryClose: LogEntryType = "close"


@dataclasses.dataclass
class LogEntry:
    """
    LogEntry is a struct-like class for serializing log entries.

    Note: the `data` field below is serialized to 'DataJSON'.
    """

    type: LogEntryType = LogEntryUnknown
    time: datetime.datetime = ZeroTime
    uuid: uuidpkg.UUID = ZeroUUID
    summary: str = ""
    data: dict = dataclasses.field(default_factory=dict)

    @staticmethod
    def from_json(j: dict) -> "LogEntry":
        entry = LogEntry()
        entry.unmarshal_json(j)
        return entry

    def unmarshal_json(self, j: dict):
        if "Type" in j:
            self.type = LogEntryType(j["Type"])
        if "Time" in j:
            self.time = isoparse(j["Time"])
        if "UUID" in j:
            self.uuid = uuidpkg.UUID(j["UUID"])
        if "Summary" in j:
            self.summary = j["Summary"]
        if "DataJSON" in j:
            self.data = json.loads(j["DataJSON"])

    def to_json(self) -> dict:
        j: dict = {}
        self.marshal_json(j)
        return j

    def marshal_json(self, j: dict):
        j["Type"] = self.type
        j["Time"] = self.time.isoformat()
        j["UUID"] = str(self.uuid)
        j["Summary"] = self.summary
        j["DataJSON"] = json.dumps(self.data)



import pytest
import io
from typing import IO

from srspy import runs

print("THIS IS SMOKE TEST 2: IT TESTS basics")

# Sundry basic tests
assert runs.DEFAULT_LOG_DIR

assert runs.now_str()
assert " " not in runs.now_str()  # doesn't have spaces

assert runs.File
assert runs.FS
assert runs.LocalFS
assert runs.LocalFS.open("/dev/null")  # basic open


# Some stubs for the file system stuff.

## File stub


class StubFile(object):
    mode: str
    closed: bool = False
    flushed: bool = True
    buffer: IO

    def __init__(self, mode: str):
        if "w" not in mode:
            raise ValueError("StubFile: mode not writable")

        self.mode = mode
        self.buffer = io.BytesIO()

    def write(self, bs: bytes):
        if self.closed:
            raise Exception("StubFile: write on closed file")

        self.buffer.write(bs)
        self.Flushed = False

    def flush(self):
        # no op
        self.flushed = True

    def close(self):
        self.closed = True


with pytest.raises(ValueError):
    StubFile("asdf")  # mode does not have w

with pytest.raises(Exception):
    f = StubFile("wb")
    assert f.closed
    f.write()  # file closed

assert StubFile("wb").flushed

f = StubFile("wb")
f.write(b"asdf")


## FS stub


class StubFS(object):
    files: dict

    def __init__(self):
        self.files = {}

    def open(self, path, mode):
        file = StubFile(mode)
        self.files[path] = file
        return file


fs = StubFS()
assert len(fs.files) == 0
fs.open("one", "wb")
fs.open("two", "wb")
assert len(fs.files) == 2

## Test RunTrace

with pytest.raises(ValueError):
    runs.RunTrace()  # lacks name

## Use the stubbed file system
fs = StubFS()
r = runs.RunTrace(name="test", fs=fs)
r.log(summary="this is a test", data={"metric": 100})
r.flush(summary="this will flush right after the write", data={"metric": 101})
r.close()

assert len(fs.files) == 1
fname = list(fs.files.keys())[0]
assert " " not in fname
file = fs.files[fname]
assert file.flushed
assert file.closed
# example value (we don't compare on this because the times and UUIDs change)
#  b'{"Type": "log", "Time": "2023-06-07T14:44:08.774714", "UUID": "bd998f0b-04c8-479a-a480-8b0a2f451897", "Summary": "test 2023-06-07_14:44:08.774712", "DataJSON": "{}"}\n{"Type": "log", "Time": "2023-06-07T14:44:08.774745", "UUID": "60af77d8-f024-4988-aa50-36699360be2b", "Summary": "this is a test", "DataJSON": "{\\"metric\\": 100}"}\n{"Type": "log", "Time": "2023-06-07T14:44:08.774757", "UUID": "8e91476d-9828-482e-8295-dd90b3626737", "Summary": "this will flush write after the write", "DataJSON": "{\\"metric\\": 101}"}\n{"Type": "close", "Time": "2023-06-07T14:44:08.774769", "UUID": "f35bbb42-5151-44a2-bb86-56d7d0b7a961", "Summary": "", "DataJSON": "{}"}\n'  # noqa: E501
val = str(file.buffer.getvalue())
assert "test" in val
assert "this is a test" in val
assert "this will flush right after" in val
assert "metric" in val

## Use the actual local file system
r = runs.RunTrace(name="test", log_dir="/tmp")
r.log(summary="this is a test", data={"metric": 100})
r.flush(summary="this will flush right after the write", data={"metric": 101})
r.close()



import datetime
import uuid

from srspy import records

print("THIS IS SMOKE TEST 1: IT TESTS records.py")

# Test zero types
assert records.ZeroTime
assert records.ZeroUUID

# Test LogEntryType enum
assert records.LogEntryUnknown != records.LogEntryLog
assert records.LogEntryLog != records.LogEntryClose
assert records.LogEntryClose != records.LogEntryUnknown

# Test LogEntry

## Test an empty one basics...
entry = records.LogEntry()
assert entry.type == records.LogEntryUnknown
assert entry.time == records.ZeroTime
assert entry.uuid == records.ZeroUUID
assert entry.summary == ""
assert entry.data == {}

json_dict = {
    "Type": "",
    "Time": "0001-01-01T00:00:00+00:00",
    "UUID": "00000000-0000-0000-0000-000000000000",
    "Summary": "",
    "DataJSON": "{}",
}

assert entry.to_json() == json_dict

d = {}
entry.marshal_json(d)
assert d == json_dict

x = records.LogEntry()
x.unmarshal_json(json_dict)
assert x == entry

y = records.LogEntry.from_json(json_dict)
assert y == entry

## Try one with some fake data

entry = records.LogEntry(
    type=records.LogEntryLog,
    time=datetime.datetime(2023, 1, 1, 3, 3, tzinfo=datetime.timezone.utc),
    uuid=uuid.UUID("47d84f08-eac2-42b1-9569-740c88f4069a"),
    summary="This is a summary! \n with new lines",
    data={
        "here is": "some data",
        "and that": {
            "is neat": 123,
        },
    },
)


json_dict = {
    "Type": "log",
    "Time": "2023-01-01T03:03:00+00:00",
    "UUID": "47d84f08-eac2-42b1-9569-740c88f4069a",
    "Summary": "This is a summary! \n with new lines",
    "DataJSON": '{"here is": "some data", "and that": {"is neat": 123}}',
}

assert entry.to_json() == json_dict

loaded = records.LogEntry.from_json(json_dict)
assert loaded == entry



# simple logging for experiment runs; see tests/*.py for examples
from typing import IO, Protocol
import datetime
import json
import os
import uuid

from .records import LogEntry, LogEntryType, LogEntryLog, LogEntryClose

# default dir to use; override via RunTrace(..., log_dir=<here>, ...)
DEFAULT_LOG_DIR: str = "./runs/"


def now_str():
    """
    A helper function to get the current time as a string that
    we can use as a path (e.g., it has no spaces).

    This function is a helper to `RunTrace.__init__` below
    """
    return str(datetime.datetime.now()).replace(" ", "_")


# A type-hint interface for a file system.
class FS(Protocol):
    def open(self, name: str, mode: str = "wb") -> IO[bytes]:
        ...


# A type-hint interface for a file.
class File(Protocol):
    def write(self, b: bytes):
        ...

    def flush(self):
        ...

    def close(self):
        ...


# An implementation of `FS` for the local file system.
class LocalFS(object):
    @staticmethod
    def open(name: str, mode: str = "wb"):
        return open(name, mode)


class RunTrace(object):
    """
    RunTrace is a class for logging information about an experiment.
    """

    _fs: FS
    closed: bool = False
    name: str
    log_file: File  # e.g. local file, file-like, spin file

    def __init__(
        self,
        name: str = "",
        log_dir: str = "",
        data: dict = {},
        fs: FS = LocalFS,
        verbose=False,
    ):
        if name == "":
            raise ValueError("RunTrace.init: must provide keyword argument `name`")
        self.name = name

        if log_dir == "":
            log_dir = DEFAULT_LOG_DIR
        self.log_dir = log_dir

        self.log_file_path = os.path.join(self.log_dir, f"{name}_{now_str()}.json")

        self.verbose = verbose
        if verbose:
            print(f"RunTrace.init at path: {self.log_file_path}")

        self._fs = fs
        self.log_file = self._fs.open(self.log_file_path, "wb")
        self.log(summary=f"{name} {now_str()}", data=data)

    def log(self, summary: str = "", data: dict = {}, type: LogEntryType = LogEntryLog):
        if self.closed:
            raise Exception("RunTrace.log: called on closed log")

        s = json.dumps(
            LogEntry(
                type=type,
                time=datetime.datetime.now(),
                uuid=uuid.uuid4(),
                summary=summary,
                data=data,
            ).to_json(),
        )
        self.log_file.write(bytes(s, "utf-8"))
        self.log_file.write(b"\n")

    def flush(self, summary: str = "", data: dict = {}):
        if self.closed:
            raise Exception("RunTrace.flush: called on closed log")

        if summary != "" or data != {}:
            self.log(summary=summary, data=data)

        self.log_file.flush()

    def close(self, summary: str = "", data: dict = {}):
        if self.closed:
            raise Exception("RunTrace.close: called on closed log")

        if summary != "" or data != {}:
            self.log(summary=summary, data=data, type=LogEntryClose)

        self.log_file.flush()
        self.log_file.close()
        self.closed = True
